export class Map {
		
	private container: any;
	private div: any;
	private canvas: HTMLCanvasElement | undefined;
	private ctx: CanvasRenderingContext2D | undefined;
	private initialised = false;
	private unit = 10;
	private xCount = 0;
	private yCount = 0;

	private mapData: any;
	private clientW: number;
	private clientH: number;
	private margin = 20;
	private scale: number;
	
	private clickEvent;
	private selectedSpaces = [];
	
	constructor(container: any) {
		this.container = container;
	}
	
	show(data?: any): void {
		this.mapData = data;
		
		this.initPanel(data);
		
		this.draw(data);
	}
	
	addClickEvent(evt: any): void {
		this.clickEvent = evt;
	}
	
	private initPanel(data: any): void{
		this.clientW = this.container.clientWidth;
		this.clientH = this.container.clientHeight;
		
		this.div = document.createElement('div');
		this.setScale(0);
		this.container.appendChild(this.div);
				
		const canvas = document.createElement('canvas');		
		canvas.width = this.clientW; // 实际渲染像素
		canvas.height = this.clientH; // 实际渲染像素
		canvas.style.width = this.clientW + 'px'; // 控制显示大小
		canvas.style.height = this.clientH + 'px'; // 控制显示大小
		canvas.style.pointerEvents = 'none';
		canvas.style.position = 'relative';
								
		this.div.appendChild(canvas);
		this.canvas = canvas;
		
		this.container.addEventListener("scroll", ()=>{this.redrawOnScroll(this)});
		
		this.container.addEventListener('click', event=>{this.doClick(event)}, true);
				
		const ctx = canvas.getContext('2d');
		if (ctx) {
			ctx.translate(0.5, 0.5);
			this.ctx = ctx;
		} else {
			throw new Error('Canvas 2D context not found, please check it is running in Browser environment.');
		}
		
		this.initialised = true;
	}
	
	private setScale(scale: number){
		if(scale == 0)
			scale = Math.min((this.clientW - this.margin*2) / this.mapData.edge.xMax.x, (this.clientH - this.margin*2) / this.mapData.edge.yMax.y);
			
		this.scale = scale;
		this.div.style.width = Math.round(this.mapData.edge.xMax.x * this.scale + 2 * this.margin) + 'px';
		this.div.style.height = Math.round(this.mapData.edge.yMax.y * this.scale + 2 * this.margin) + 'px';
	}
	
	scaleTo(scale: number):void{
		const baseY = this.clientH / 2;
		const baseX = this.clientW / 2;
		const yRatio = this.container.scrollTop == 0 ? 0 : (this.container.scrollTop + baseY) / this.div.clientHeight;
		const xRatio = this.container.scrollLeft == 0 ? 0 : (this.container.scrollLeft + baseX) / this.div.clientWidth;
		
		this.setScale(scale);
		
		var top = 0
		var left = 0;
		if(yRatio > 0) {
			top = this.div.clientHeight * yRatio - baseY;
			if(top > 0){
				const tmp = this.div.clientHeight - (top + this.clientH);
				if(tmp < 0)
					top += tmp;
			}
		}
		if(xRatio > 0) {
			left = this.div.clientWidth * xRatio - baseX;	
			if(left > 0){
				const tmp = this.div.clientWidth - (left + this.clientW);
				if(tmp < 0)
					left += tmp;
			}
		}
			
		this.container.scrollTop = top > 0 ? top : 0;
		this.container.scrollLeft = left > 0 ? left : 0;
		
		this.redrawOnScroll(this);
	}
	
	private paintBg(){
		for(var x = this.unit; x < this.canvas.width; x += this.unit){
			this.xCount++;

			if(this.xCount % 5 == 0)
				this.ctx.strokeStyle = "#f1f1f1";
			else
				this.ctx.strokeStyle = "#f7f7f7";
								
			this.ctx.beginPath();
			this.ctx.moveTo(x, 0);
			this.ctx.lineTo(x, this.canvas.height);
			this.ctx.stroke();
		}
		for(var y = this.unit; y < this.canvas.height; y += this.unit){
			this.yCount++;
			
			if(this.yCount % 5 == 0)
				this.ctx.strokeStyle = "#f4f4f4";
			else
				this.ctx.strokeStyle = "#fafafa";
				
			this.ctx.beginPath();
			this.ctx.moveTo(0, y);
			this.ctx.lineTo(this.canvas.width, y);
			this.ctx.stroke();
		}		
	}
	
	private redrawOnScroll(_this) {
		_this.ctx.clearRect(-1, -1, _this.canvas.width + 1, _this.canvas.height + 1);  
				
		const top = _this.container.scrollTop;
		const left = _this.container.scrollLeft;
		_this.canvas.style.top = top + 'px';
		_this.canvas.style.left = left + 'px';
		
		_this.draw(_this.mapData);
	}

	private draw(data: any):void {
		this.paintBg();
		
		for(const w of data.walls){
			this.ctx.strokeStyle = "#cccccc";
			this.paintPath(w, true);
			
			// this.ctx.strokeStyle = "green";
			// if(w.location != null){
			// 	this.paintPath(w.location);
			// }
			
			// var p = w.path[0];
			// this.drawText(w.id, p);
		}
		
		this.ctx.strokeStyle = "#000000";
		for(const pillar of data.pillars){
			if(pillar.roomBoundary)
				this.paintPath(pillar, true);
		}
		
		this.ctx.strokeStyle = "#b65b00";
		for(const d of data.doors){
			this.paintPath(d, true);
		}
		
		this.ctx.strokeStyle = "#0000ff";
		this.ctx.setLineDash([2,4]);
		for(const vw of data.vws){
			this.paintPath(vw, true);
		}
		this.ctx.setLineDash([]);
		
		const fillColor = "rgba(0, 0, 255, 0.1)";
		const fillColors = ["rgba(255, 0, 0, 0.3)","rgba(215, 0, 0, 0.3)","rgba(175, 0, 0, 0.3)","rgba(135, 0, 0, 0.3)","rgba(95, 0, 0, 0.3)","rgba(55, 0, 0, 0.3)","rgba(15, 0, 0, 0.3)","rgba(0, 255, 0, 0.3)","rgba(0, 215, 0, 0.3)","rgba(0, 175, 0, 0.3)","rgba(0, 135, 0, 0.3)","rgba(0, 95, 0, 0.3)","rgba(0, 55, 0, 0.3)","rgba(0, 15, 0, 0.3)","rgba(0, 0, 255, 0.3)","rgba(0, 0, 215, 0.3)","rgba(0, 0, 175, 0.3)","rgba(0, 0, 135, 0.3)","rgba(0, 0, 95, 0.3)","rgba(0, 0, 55, 0.3)","rgba(0, 0, 15, 0.3)"];
		this.ctx.strokeStyle = "#00BB00";
		var idx = 0;
		for(const sp of data.spaces){
			if(!sp.inner || (!sp.virtual && sp.divided))
				continue;
			
			this.paintSpaceName(sp);
			
			// if(!sp.virtual){
			// 	if(sp.inner && !sp.divided) {
			// 		this.ctx.fillStyle = fillColor;
			// 		this.fillPath(sp);	
			// 	}
			// } else {
			// 	this.ctx.fillStyle = fillColors[idx++];
			// 	this.fillPath(sp);
			// }
			
			
			//idx++;
		}
		
		// for(const idx of [29]){
		// 	var sp = data.spaces[idx];
		// 	this.fillPath(sp);
		// 	console.log(sp.path);	
		// }
	}
	
	private paintPath(element: any, close?: boolean): void{
		// if(element.id != '10149')
		// 	return;
				
		if(!element.path || element.path.length == 0)
			return;
		
		// var p = element.path[0];
		// this.drawText(element.id, p);
		
		this.ctx.beginPath();
		
		this.movePath(element.path); //element.handledPath || element.path
		if(close)
			this.ctx.closePath();
			
		this.ctx.stroke();
			
		// const spLines = element.spaceLines;
		// if(!spLines)
		// 	return;
		// for(const l of spLines){
		// 	this.ctx.moveTo(this.transX(l.point1.x), this.transY(l.point1.y));
		// 	this.ctx.lineTo(this.transX(l.point2.x), this.transY(l.point2.y));	
		// }
		// this.ctx.stroke();
	}
	
	private fillPath(element: any): void{
		this.ctx.beginPath();
		
		this.movePath(element.path, true);
		
		if(element.holes){
			for(const id of element.holes){
				const hole = this.getSpace(id);
				this.movePath([...hole.path].reverse(), true);
			}
		}

		this.ctx.fill();
		
		// if(element.holes){
		// 	this.ctx.beginPath();
		// 	for(const id of element.holes){
		// 		const hole = this.getSpace(id);
		// 		this.movePath(hole.path, true);
		// 	}
		// 	this.ctx.stroke();
		// }
		
		// this.ctx.beginPath();
		// this.ctx.arc(this.transX(element.interior.x), this.transY(element.interior.y), 2, 0, 2 * Math.PI);
		// this.ctx.stroke();
		
	}
	
	private paintSpaceName(space: any): void{
		if(space.area > 1000 * 1000 * 4)
			this.drawText(space.id, space.interior, true);
	}
	
	private movePath(path: any[], manualClose?: boolean): void{
		var pre = null;
		for(const p of path){
			if(!pre)
				this.ctx.moveTo(this.transX(p.x), this.transY(p.y));
			else
				this.ctx.lineTo(this.transX(p.x), this.transY(p.y));
			pre = p;
		}
		if(manualClose){
			if(pre.x != path[0].x || pre.y != path[0].y)
				this.ctx.lineTo(this.transX(path[0].x), this.transY(path[0].y));
		}
	}
	
	private getSpace(id: string): any {
		for(const sp of this.mapData.spaces){
			if(sp.id == id)
				return sp;
		}
	}

	private doClick(event): void {
		const left = 200;
		const top = 33;
		var x = event.clientX - left + this.container.scrollLeft - this.margin;
		var y = event.clientY - top + this.container.scrollTop - this.margin;
		x = x / this.scale;
		y = y / this.scale;
		//alert(x + "," + y);
		this.clickEvent.onClick(x, y);
	}
	
	public selectSpace(spId): void {
		this.redrawOnScroll(this);
		
		this.selectedSpaces = [];
		if(spId){
			this.ctx.fillStyle = "rgba(255, 0, 0, 0.1)";
			const sp = this.getSpace(spId);
			this.fillPath(sp);
			this.selectedSpaces.push(sp);
			
			this.ctx.fillStyle = "rgba(0, 0, 255, 0.03)";
			if(sp.nearby){
				for(const id of sp.nearby){
					const nsp = this.getSpace(id);
					this.fillPath(nsp);
				}
			}
		}
	}
	
	public update(data: any): void {
		this.mapData = data;
		this.redrawOnScroll(this);
	}

	private transX(x: number): number {
		return Math.round(x * this.scale + this.margin - this.container.scrollLeft);
	}
	
	private transY(y: number): number {
		return Math.round(y * this.scale + this.margin - this.container.scrollTop);
	}
	
	private drawText(text: string, point: any, center?:boolean):void{
		const fontSize = 10;
		this.ctx.font = "lighter " + fontSize + "px Microsoft YaHei";
		this.ctx.textBaseline = "bottom";
		
		const bakColor = this.ctx.strokeStyle;
		this.ctx.strokeStyle = "#000000";
		
		var xOffset = 0;
		var yOffset = 0;
		if(center) {
			xOffset = this.ctx.measureText(text).width / 2 * -1;
			yOffset = fontSize / 2;		
		}
		this.ctx.strokeText(text, this.transX(point.x) + xOffset, this.transY(point.y) + yOffset);
		
		this.ctx.strokeStyle = bakColor;
	}

}