import {Point} from './UIUtils';

/**
 * 系统图相关模型
 * @author zhaoyk
 */
export class Template {
	
	diagramType:string;
	
	id:string;
	
	name:string;
	
	code:string;
	
	frame:Container;
	
	mainPipes: Array<MainPipe>;
	
	scatteredContainers: Array<Container>;

	static getContainerById(id: string, template: Template): Container {
		return Container.getContainerById(id, template.frame);
	}
	
	static getMainPipeById(id: string, template: Template): MainPipe {
		if(template.mainPipes){
			for(const mp of template.mainPipes){
				if(mp.id == id)
					return mp;
			}
		}
		return null;
	}
		
	static getElById(id: string, template: Template): any {
		var el: any = Template.getContainerById(id, template);
		if(el == null)
			el = Template.getMainPipeById(id, template);
		return el;
	}

	static getContainers(template: Template): Array<Container> {
		const result = new Array<Container>();
		Template.doGetContainers(template.frame, result);
		return result;
	}
	
	private static doGetContainers(con: Container, result: Array<Container>){
		result.push(con);
		if(con.children && con.children.length > 0){
			con.children.forEach(comp => {
				if(comp.compType == 'container')
					Template.doGetContainers(<Container>comp, result);
			});
		}
	}
	
	static findLeaderEls(el: any, template: Template): Array<any> {
		const list = new Array<any>();
		for(const con of Template.getContainers(template)) {
			if(con != el && Container.isEquipBox(con)) {
				if(Template.canBeLeader(el, con, template))
					list.push(con);
			}
		}
		for(const mp of template.mainPipes) {
			if(mp != el && mp.bindEquipment) {
				if(Template.canBeLeader(el, mp, template))
					list.push(mp);
			}
		}
		return list;
	}
	
	private static canBeLeader(el: any, leader: any, template: Template): boolean {
		const pdc = Template.getParentDynCon(el, template);
		const lPdc = Template.getParentDynCon(leader, template);
		
		if(pdc == lPdc){
			if(leader.dynTarget != null || el.dynTarget == null) {
				if(el.compType != leader.compType){
					return el.compType == 'mainPipe' ? Template.isRelatedWith(<MainPipe>el, <Container>leader, template) : Template.isRelatedWith(<MainPipe>leader, <Container>el, template);
				} else
					return true;	
			}
		} else {
			if(el.compType == leader.compType) {
				if(!lPdc)
					return true;
			}
			if(pdc && lPdc)
				return Container.isParentOf(lPdc, pdc);	
		}
		return false;
	}
	
	private static getParentDynCon(el: any, template: Template): Container {
		var p: Container;
		if(el.compType == 'container') {
			p = (<Container>el).parent;
			while(p != null) {
				if(p.dynGroup)
					return p;
				p = p.parent;
			}
		} else if(el.compType == 'mainPipe') {
			const relConIds = (<MainPipe>el).relatedContainers;
			if(relConIds){
				for(const conId of relConIds) {
					const relCon = Template.getContainerById(conId, template);
					if(relCon) {
						if(relCon.dynGroup)
							return relCon;
						p = Template.getParentDynCon(relCon, template);
						if(p)
							return p;
					}
				}
			}
		}
		return null;
	}
	
	static isRelatedWith(mainPipe: MainPipe, con: Container, template: Template): boolean {
		if(mainPipe.relatedContainers){
			for(const conId of mainPipe.relatedContainers) {
				const relCon = Template.getContainerById(conId, template);
				if(relCon) {
					if(relCon == con || Container.isParentOf(relCon, con))
						return true;
				}
			}
		}
		return false;
	}
	
	static clearParent(template: Template){
		Template.clearConParent(template.frame);
	}
	
	private static clearConParent(con:Container){
		con.parent = null;
		con.children.forEach(item => {
			item.parent = null;
			
			if(item.compType == 'container') 
				Template.clearConParent(<Container>item);
		});
	}
	
}

export class Comp {

	id:string;
	
	name:string;
	
	location:Point;
	
	size:Point;
	
	position: any;
	
	compType: string;
	
	parent: Container;
				
}

export class Container extends Comp {
	
	remark:string;
	
	children: Array<Comp>;

	layout: any;
	
	equipmentTypes: Array<string>;

	dataFilter: DataFilter;
	
	equipPack: EquipPack;
	
	dynGroup: DynGroup;
	
	props: Object;
	
	dynTarget: string;
	
	static isParentOf(parent: Container, comp: Comp): boolean{
		var p = comp.parent;
		while (p != null) {
			if(p == parent){
				return true;
			}
			p = p.parent;
		}
		return false;
	}
	
	static isEquipBox(con: Container): boolean {
		return con && con.equipmentTypes != null && con.equipmentTypes.length > 0;
	}
	
	static getEquipBoxes(con: Container, arr?: Array<Container>){
		if(!arr)
			arr = new Array<Container>();
		if(Container.isEquipBox(con))
			arr.push(con);
		else if(con.children && con.children.length > 0) {
			for(const item of con.children) {
				Container.getEquipBoxes(<Container>item, arr);
			}
		}
		return arr;
	}
	
	static findSubCons(con: Container, arr?: Array<Container>) {
		if(!arr)
			arr = new Array<Container>();
		if(con.children && con.children.length > 0) {
			for(const item of con.children) {
				arr.push(<Container>item);
				Container.findSubCons(<Container>item, arr);
			}
		}
		return arr;
	}
	
	static getContainerById(id: string, con: Container): Container {
		if(con.id == id)
			return con;
		if(con.children && con.children.length > 0){
			for(const c of con.children) {
				if(c.compType == 'container') {
					const rtn = Container.getContainerById(id, <Container>c);
					if(rtn != null)
						return rtn;	
				}
			}
		}
		return null;
	}
	
}

export class DataFilter {
	
	type: number;
	
	object: string;
	
	operation: string;
	
	value: string;
	
	items: Array<DataFilter>;
	
	not: boolean;
	
}

export class EquipPack {
	
	packByType: boolean;
	
	legendId: string;
	
	packName: string;

}

export class DynGroup {

	dynSource: string;
	
	dynSourceType: string;
	
	labelPosition: string;

}

export class MainPipe {
	
	id:string;
	
	name:string;
	
	remark:string;
	
	type:string;
	
	bindEquipment:boolean;
	
	path: any[];
	
	relatedContainers: Array<string>;
	
	locationPath: Array<Array<Point>>;
			
	dynTarget: string;
	
}

export class Legend {

	id:string;
	
	code:string;

	name:string;

	remark:string;

	width:number;

	height:number;

	equipmentTypes: Array<string>;

	diagramTypes: Array<string>;

	anchors: Array<Anchor>;
	
	dataFilter: DataFilter;
	
	outline: Array<Array<Point>>;

	infos: Array<string>;
	
	iconId: string;
	
	static getPointsForAnchor(legend: Legend, location: Point): Array<Point4Anchor> {
		const size = new Point(legend.width, legend.height);
		
		const points = new Array<Point4Anchor>();
		points.push(new Point4Anchor(new Point(location.x, location.y), Point4Anchor.TL));
		points.push(new Point4Anchor(new Point(location.x + size.x, location.y), Point4Anchor.TR));
		points.push(new Point4Anchor(new Point(location.x + size.x, location.y + size.y), Point4Anchor.BR));
		points.push(new Point4Anchor(new Point(location.x, location.y + size.y), Point4Anchor.BL));
		
		const part = 6;
		var len = size.x / part;
		for (var i = 1; i < part; i++) {
			points.push(new Point4Anchor(new Point(location.x + len * i, location.y), "T" + i));
			points.push(new Point4Anchor(new Point(location.x + len * i, location.y + size.y), "B" + i));
		}
		len = size.y / part;
		for (var i = 1; i < part; i++) {
			points.push(new Point4Anchor(new Point(location.x, location.y + len * i), "L" + i));
			points.push(new Point4Anchor(new Point(location.x + size.x, location.y + len * i), "R" + i));
		}
		
		if(legend.anchors) {
			for(const anchor of legend.anchors) {
				for(const p of points) {
					if(p.code == anchor.code) {
						p.anchor = anchor;
						p.selected = true;
						break;
					}
				}
			}	
		}
		
		return points;
	}
	
}

export class Anchor {
	
	code: string;
	
	name: string;
	
	type: string;
	
	acceptRelations: Array<string>

	toEquipmentTypes: Array<string>
	
	toDataFilter: DataFilter;

	attachToOutline: boolean = true;
	
	multiple: boolean;
	
	shape: string;


	xOffset: number;
	
	yOffset: number;
	
}

export class Point4Anchor {
	
	static TL = "TL";
	static TR = "TR";
	static BR = "BR";
	static BL = "BL";
	
	point: Point;
	code: string;
	selected: boolean;
	anchor: Anchor;
	
	constructor(point: Point, code: string) {
		this.point = point;
		this.code = code;
	}
	
}


export class Diagram {
	
	id:string;
	
	name: string;
	
	type: string;
	
	system: string;
	
	templateId: string;
	
	nodes: Array<DiagramNode>;
	
	lines: Array<Line>;
	
	template: Template;
	
}

export class DiagramNode extends Comp {
	
	containerId: string;
	
	layoutIndex: number;
	
	dataObject: Object;
	
}

export class EquipmentNode extends DiagramNode {
	
	dataObjectId: string;
	
	objClassCode: string;

	legendId: string;

	legend: Legend;
	
	anchorLocations: Object;
	
	label: Label;
	
}

export class Label extends Comp {
	
	content: string;
	
}

export class Line {
	
	locationPath: Array<Point>;
	
	flag: string;
	
}