import {DataFilter} from './DiagramModel';
/**
 * ui操作的基础工具
 * @author zhaoyk
 */

export class Point {
	
	x: number;
	y: number;
	
	constructor(pX: number, pY: number) {
		this.x = pX;
		this.y = pY;
	}
	
}

export class BackgroundPaniter {
		
	private canvas: HTMLCanvasElement;
	
	private ctx: CanvasRenderingContext2D;
	
	private unit = 10;
	
	constructor(canvas: HTMLCanvasElement, ctx: CanvasRenderingContext2D) {
		this.canvas = canvas;
		this.ctx = ctx;
	}
	
	paint(){
		var xCount = 0;
		var yCount = 0;
		this.ctx.lineWidth = 1;
		for(var x = this.unit; ; x += this.unit){
			if(x > this.canvas.width)
				break;
				
			xCount++;
			if(xCount % 5 == 0)
				this.ctx.strokeStyle = "#f1f1f1";
			else
				this.ctx.strokeStyle = "#f7f7f7";
								
			this.ctx.beginPath();
			this.ctx.moveTo(x, 0);
			this.ctx.lineTo(x, this.canvas.height);
			this.ctx.stroke();
		}
		for(var y = this.unit; ; y += this.unit){
			if(y > this.canvas.height)
				break;
				
			yCount++;
			if(yCount % 5 == 0)
				this.ctx.strokeStyle = "#f4f4f4";
			else
				this.ctx.strokeStyle = "#fafafa";
				
			this.ctx.beginPath();
			this.ctx.moveTo(0, y);
			this.ctx.lineTo(this.canvas.width, y);
			this.ctx.stroke();
		}		
	}
	
}

export class ViewTool {
	
	static paintPoint(ctx: CanvasRenderingContext2D, point:Point, color:string, fill: boolean, r?: number): void {
		if(!r)
			r = 2;
		if(fill) {
			ctx.fillStyle = color;
			ctx.arc(point.x, point.y, r, 0, Math.PI * 2);			
			ctx.fill();
		} else {
			ctx.strokeStyle = color;
			ctx.arc(point.x, point.y, r, 0, Math.PI * 2);
			ctx.stroke();
		}
	}
	
	static drawText(ctx: CanvasRenderingContext2D, content: string, location: Point, size: Point):void{
		const fontSize = 10;
		ctx.font = "lighter " + fontSize + "px Microsoft YaHei";
		ctx.textBaseline = "bottom";
		
		var rows = [];
		var tmp = "";
		for(var i = 0; i < content.length; i++) {
			var nl = content[i] == '\n';
			if(!nl){
				tmp += content[i];
				nl = ctx.measureText(tmp).width >= size.x - 20;
			}
			if(nl) {
				rows.push(tmp);
				tmp = "";
			}
		}
		if(tmp.length > 0)
			rows.push(tmp);
		
		for(var i = 0; i < rows.length; i++) {
			ctx.strokeText(rows[i], location.x, location.y + ((i + 1) * (fontSize + 4)));
		}
	}
	
	static pointToStr(point: Point): string {
		return '(x: ' + point.x + ', y: ' + point.y + ')';
	}
	
	static pointsToStr(points: Array<Point>): string {
		var str = '';
		if(points && points.length){
			points.forEach(item => {
				if(str.length > 0)
					str += ' , ';
				str += ViewTool.pointToStr(item);
			});
		}
		return '[' + str + ']';
	}
	
	static filterToStr(filter: DataFilter) {
		var info = '';
		if(filter) {
			info += ' :'
			if(filter.not)
				info += '!';
			if(filter.type == 1)
				info += filter.value;
			else if(filter.type == 9)
				info += '关系过滤';
			else {
				info += '组合过滤';
			}
		}
		return info;
	}
	
}